/*
 * Copyright (c) 1983 Eric P. Allman
 * Copyright (c) 1988 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted provided
 * that: (1) source distributions retain this entire copyright notice and
 * comment, and (2) distributions including binaries display the following
 * acknowledgement:  ``This product includes software developed by the
 * University of California, Berkeley and its contributors'' in the
 * documentation or other materials provided with the distribution and in
 * all advertising materials mentioning features or use of this software.
 * Neither the name of the University nor the names of its contributors may
 * be used to endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <errno.h>
#include "sendmail.h"

#ifndef lint
#ifdef DAEMON
static char sccsid[] = "@(#)daemon.c	5.36 (Berkeley) 6/1/90 (with daemon mode)";
static char rcsid[] = "$Header: /wrl/gen/gwsrc/sendmail/ida-5.65/src/RCS/daemon.c,v 1.9 1991/10/28 04:50:45 vixie Exp $ (with daemon mode)";
#else
static char sccsid[] = "@(#)daemon.c	5.36 (Berkeley) 6/1/90 (without daemon mode)";
static char rcsid[] = "$Header: /wrl/gen/gwsrc/sendmail/ida-5.65/src/RCS/daemon.c,v 1.9 1991/10/28 04:50:45 vixie Exp $ (without daemon mode)";
#endif
#endif /* not lint */

int la;	/* load average */

#ifdef DAEMON

# include <netdb.h>
# include <sys/signal.h>
# include <sys/wait.h>
# include <sys/time.h>
# include <sys/resource.h>
# ifdef IDA
#  include <sys/file.h>
#  include <sys/types.h>
#  include <sys/stat.h>
# endif
# ifdef DECWRL
#  if defined(MININODES) && defined(MINBLOCKS)
#   if defined(sun)
#    include <sys/vfs.h>
#   endif
#   if defined(ultrix)
#    include <sys/param.h>
#    include <sys/mount.h>
#   endif ultrix
#  endif sun
# endif DECWRL

/*
**  DAEMON.C -- routines to use when running as a daemon.
**
**	This entire file is highly dependent on the 4.2 BSD
**	interprocess communication primitives.  No attempt has
**	been made to make this file portable to Version 7,
**	Version 6, MPX files, etc.  If you should try such a
**	thing yourself, I recommend chucking the entire file
**	and starting from scratch.  Basic semantics are:
**
**	getrequests()
**		Opens a port and initiates a connection.
**		Returns in a child.  Must set InChannel and
**		OutChannel appropriately.
**	clrdaemon()
**		Close any open files associated with getting
**		the connection; this is used when running the queue,
**		etc., to avoid having extra file descriptors during
**		the queue run and to avoid confusing the network
**		code (if it cares).
**	makeconnection(host, port, outfile, infile)
**		Make a connection to the named host on the given
**		port.  Set *outfile and *infile to the files
**		appropriate for communication.  Returns zero on
**		success, else an exit status describing the
**		error.
**	maphostname(hbuf, hbufsize)
**		Convert the entry in hbuf into a canonical form.  It
**		may not be larger than hbufsize.
**
**	mapinit(c)
**		Open and initialize a dbm database.  Reopen if our current
**		file descriptor is out of date.
**
**	mapkey(c, key, argval, argsiz)
**		Search a database for a match to the given key, sprintf'ing
**		the argument through the result if found.
*/
/*
**  GETREQUESTS -- open mail IPC port and get requests.
**
**	Parameters:
**		none.
**
**	Returns:
**		none.
**
**	Side Effects:
**		Waits until some interesting activity occurs.  When
**		it does, a child is created to process it, and the
**		parent waits for completion.  Return from this
**		routine is always in the child.  The file pointers
**		"InChannel" and "OutChannel" should be set to point
**		to the communication channel.
*/

struct sockaddr_in	SendmailAddress;/* internet address of sendmail */

int	DaemonSocket	= -1;		/* fd describing socket */
#ifndef DECWRL
char	*NetName;			/* name of home (local?) network */
#endif DECWRL

getrequests()
{
	int t;
	register struct servent *sp;
	int on = 1;
	extern reapchild();

	/*
	**  Set up the address for the mailer.
	*/

	sp = getservbyname("smtp", "tcp");
	if (sp == NULL)
	{
		syserr("server \"smtp\" unknown");
		goto severe;
	}
	SendmailAddress.sin_family = AF_INET;
	SendmailAddress.sin_addr.s_addr = INADDR_ANY;
	SendmailAddress.sin_port = sp->s_port;

	/*
	**  Try to actually open the connection.
	*/

	if (tTd(15, 1))
		printf("getrequests: port 0x%x\n", SendmailAddress.sin_port);

	/* get a socket for the SMTP connection */
	DaemonSocket = socket(AF_INET, SOCK_STREAM, 0);
	if (DaemonSocket < 0)
	{
		/* probably another daemon already */
		syserr("getrequests: can't create socket");
	  severe:
# ifdef LOG
		if (LogLevel > 0)
			syslog(LOG_ALERT, "cannot get connection");
# endif LOG
		finis();
	}

	/* turn on network debugging? */
	if (tTd(15, 15))
		(void) setsockopt(DaemonSocket,
				  SOL_SOCKET, SO_DEBUG,
				  (char *)&on, sizeof on);

	(void) setsockopt(DaemonSocket,
			  SOL_SOCKET, SO_REUSEADDR, (char *)&on, sizeof on);
	(void) setsockopt(DaemonSocket,
			  SOL_SOCKET, SO_KEEPALIVE, (char *)&on, sizeof on);

	if (bind(DaemonSocket, &SendmailAddress, sizeof SendmailAddress) < 0)
	{
		syserr("getrequests: cannot bind");
		(void) close(DaemonSocket);
		goto severe;
	}
	if (listen(DaemonSocket, 10) < 0)
	{
		syserr("getrequests: cannot listen");
		(void) close(DaemonSocket);
		goto severe;
	}

	(void) signal(SIGCHLD, reapchild);

	if (tTd(15, 1))
		printf("getrequests: %d\n", DaemonSocket);

	for (;;)
	{
		register int pid;
		auto int lotherend;
		register enum {accepting,refusing} state = accepting;

		/* see if we are rejecting connections */
		while (rejecting_connections())
		{
			if (state == accepting) {
				(void) listen(DaemonSocket, 0);
				state = refusing;
			}
			sleep(5);
		}

		if (state == refusing) {
			(void) listen(DaemonSocket, 10);
			state = accepting;
		}

		/* wait for a connection */
		setproctitle("accepting connections");
		do
		{
			errno = 0;
			lotherend = sizeof RealHostAddr;
			t = accept(DaemonSocket, &RealHostAddr, &lotherend);
		} while (t < 0 && errno == EINTR);
		if (t < 0)
		{
			syserr("getrequests: accept");
			sleep(5);
			continue;
		}

		/*
		**  Create a subprocess to process the mail.
		*/

		if (tTd(15, 2))
			printf("getrequests: forking (fd = %d)\n", t);

		if (tTd(0, 1))
			pid = 0;
		else
			pid = fork();

		if (pid < 0)
		{
			syserr("daemon: cannot fork");
			sleep(10);
			(void) close(t);
			continue;
		}

		if (pid == 0)
		{
			extern struct hostent *gethostbyaddr();
			register struct hostent *hp;
			char buf[MAXNAME];

			/*
			**  CHILD -- return to caller.
			**	Collect verified idea of sending host.
			**	Verify calling user id if possible here.
			*/

			(void) signal(SIGCHLD, SIG_DFL);

			/* determine host name */
			hp = gethostbyaddr((char *) &RealHostAddr.sin_addr,
					   sizeof RealHostAddr.sin_addr,
					   AF_INET);
			if (hp != NULL)
				(void) strcpy(buf, hp->h_name);
			else
			{
				extern char *inet_ntoa();

				/* produce a dotted quad */
				(void) sprintf(buf, "[%s]",
					inet_ntoa(RealHostAddr.sin_addr));
			}

			/* should we check for illegal connection here? XXX */

			RealHostName = newstr(buf);

			(void) close(DaemonSocket);
			InChannel = fdopen(t, "r");
			OutChannel = fdopen(dup(t), "w");
			if (tTd(15, 2))
				printf("getreq: returning\n");
# ifdef LOG
			if (LogLevel > 11)
				syslog(LOG_DEBUG, "connected, pid=%d", getpid());
# endif LOG
			return;
		}

		/* close the port so that others will hang (for a while) */
		(void) close(t);

		/* sleep a bit before the next accept() */
		sleep(1);
	}
	/*NOTREACHED*/
}
/*
**  CLRDAEMON -- reset the daemon connection
**
**	Parameters:
**		none.
**
**	Returns:
**		none.
**
**	Side Effects:
**		releases any resources used by the passive daemon.
*/

clrdaemon()
{
	if (DaemonSocket >= 0)
		(void) close(DaemonSocket);
	DaemonSocket = -1;
}
/*
**  REJECTING_CONNECTIONS -- boolean: should connections be rejected
**
**      Parameters:
**              none.
**
**      Returns:
**              TRUE(1) if connections should be rejected or
**              FALSE(0) if connections can be accepted
**
**      Side Effects:
**              none.
**
**	Authors:
**		David Koblas (written)
**		Paul Vixie (hacked, modified, changed name)
*/

static int
rejecting_connections()
{
	extern int      RefuseLA;
	register int    la;
	char		newtitle[100];
# if defined(MININODES) && defined(MINBLOCKS)
#  if defined(sun)
	struct statfs   statbuf;
#	define INODES	f_ffree
#	define BLOCKS	f_bavail
#  endif
#  if defined(ultrix)
	struct fs_data	statbuf;
#	define INODES	fd_gfree
#	define BLOCKS	fd_bfree
#  endif
#  if !defined(INODES) || !defined(BLOCKS)
#   undef MININODES
#   undef MINBLOCKS
#  endif
# endif

	newtitle[0] = '\0';

# if defined(MININODES) && defined(MINBLOCKS)
	if (statfs(QueueDir, &statbuf) < 0) {
		sprintf("rejecting, unable to statfs(%s)", QueueDir);
        } else
	if ((statbuf.BLOCKS < MINBLOCKS) || (statbuf.INODES < MININODES)) {
		sprintf(newtitle, "rejecting, free(%s)=%di:%db",
			QueueDir, statbuf.INODES, statbuf.BLOCKS);
	} else
# endif
	if ((la = getla()) > RefuseLA) {
		sprintf(newtitle, "rejecting, loadav=%d", la);
	}

	if (newtitle[0] != '\0') {
		setproctitle(newtitle);
		syslog(LOG_INFO, newtitle);
		if (tTd(15, 1))
			puts(newtitle);
		return 1;
	}

	return 0;
}
/*
**  MAKECONNECTION -- make a connection to an SMTP socket on another machine.
**
**	Parameters:
**		host -- the name of the host.
**		port -- the port number to connect to.
**		outfile -- a pointer to a place to put the outfile
**			descriptor.
**		infile -- ditto for infile.
**
**	Returns:
**		An exit code telling whether the connection could be
**			made and if not why not.
**
**	Side Effects:
**		none.
*/

makeconnection(host, port, outfile, infile)
	char *host;
	u_short port;
	FILE **outfile;
	FILE **infile;
{
	register int i, s;
	register struct hostent *hp = (struct hostent *)NULL;
	extern char *inet_ntoa();
	int sav_errno;
#ifdef NAMED_BIND
	extern int h_errno;
#endif

	/*
	**  Set up the address for the mailer.
	**	Accept "[a.b.c.d]" syntax for host name.
	*/

#ifdef NAMED_BIND
	h_errno = 0;
#endif
	errno = 0;

	if (host[0] == '[')
	{
		long hid;
		register char *p = index(host, ']');

		if (p != NULL)
		{
			*p = '\0';
			hid = inet_addr(&host[1]);
			*p = ']';
		}
		if (p == NULL || hid == -1)
		{
			usrerr("Invalid numeric domain spec \"%s\"", host);
			return (EX_NOHOST);
		}
		SendmailAddress.sin_addr.s_addr = hid;
	}
	else
	{
		hp = gethostbyname(host);
		if (hp == NULL)
		{
#ifdef NAMED_BIND
			if (errno == ETIMEDOUT || h_errno == TRY_AGAIN)
				return (EX_TEMPFAIL);

			/* if name server is specified, assume temp fail */
			if (errno == ECONNREFUSED && UseNameServer)
				return (EX_TEMPFAIL);
#endif

			/*
			**  XXX Should look for mail forwarder record here
			**  XXX if (h_errno == NO_ADDRESS).
			*/

			return (EX_NOHOST);
		}
		bcopy(hp->h_addr, (char *) &SendmailAddress.sin_addr, hp->h_length);
		i = 1;
	}

	/*
	**  Determine the port number.
	*/

	if (port != 0)
		SendmailAddress.sin_port = htons(port);
	else
	{
		register struct servent *sp = getservbyname("smtp", "tcp");

		if (sp == NULL)
		{
			syserr("makeconnection: server \"smtp\" unknown");
			return (EX_OSFILE);
		}
		SendmailAddress.sin_port = sp->s_port;
	}

	/*
	**  Try to actually open the connection.
	*/

again:
	if (tTd(16, 1))
		printf("makeconnection (%s [%s])\n", host,
		    inet_ntoa(SendmailAddress.sin_addr.s_addr));

	s = socket(AF_INET, SOCK_STREAM, 0);
	if (s < 0)
	{
		syserr("makeconnection: no socket");
		sav_errno = errno;
		goto failure;
	}

	if (tTd(16, 1))
		printf("makeconnection: %d\n", s);

	/* turn on network debugging? */
	if (tTd(16, 14))
	{
		int on = 1;
		(void) setsockopt(DaemonSocket, SOL_SOCKET, SO_DEBUG, (char *)&on, sizeof on);
	}
	if (CurEnv->e_xfp != NULL)
		(void) fflush(CurEnv->e_xfp);		/* for debugging */
	errno = 0;					/* for debugging */
	SendmailAddress.sin_family = AF_INET;
	if (connect(s, &SendmailAddress, sizeof SendmailAddress) < 0)
	{
		sav_errno = errno;
		(void) close(s);
		if (hp && hp->h_addr_list[i])
		{
			bcopy(hp->h_addr_list[i++],
			    (char *)&SendmailAddress.sin_addr, hp->h_length);
			goto again;
		}

		/* failure, decide if temporary or not */
	failure:
		errno = sav_errno;
		switch (errno)
		{
		  case EISCONN:
		  case ETIMEDOUT:
		  case EINPROGRESS:
		  case EALREADY:
		  case EADDRINUSE:
		  case EHOSTDOWN:
		  case ENETDOWN:
		  case ENETRESET:
		  case ENOBUFS:
		  case ECONNREFUSED:
		  case ECONNRESET:
		  case EHOSTUNREACH:
		  case ENETUNREACH:
			/* there are others, I'm sure..... */
			return (EX_TEMPFAIL);

		  case EPERM:
			/* why is this happening? */
			syserr("makeconnection: funny failure, addr=%lx, port=%x",
				SendmailAddress.sin_addr.s_addr, SendmailAddress.sin_port);
			return (EX_TEMPFAIL);

		  default:
			{
				extern char *errstring();

				message(Arpa_Info, "%s", errstring(sav_errno));
				return (EX_UNAVAILABLE);
			}
		}
	}

	/* connection ok, put it into canonical form */
	*outfile = fdopen(s, "w");
	*infile = fdopen(dup(s), "r");

	return (EX_OK);
}
/*
**  MYHOSTNAME -- return the name of this host.
**
**	Parameters:
**		hostbuf -- a place to return the name of this host.
**		size -- the size of hostbuf.
**
**	Returns:
**		A list of aliases for this host.
**
**	Side Effects:
**		none.
*/

char **
myhostname(hostbuf, size)
	char hostbuf[];
	int size;
{
	extern struct hostent *gethostbyname();
	struct hostent *hp;

	if (gethostname(hostbuf, size) < 0)
	{
		(void) strcpy(hostbuf, "localhost");
	}
	hp = gethostbyname(hostbuf);
	if (hp != NULL)
	{
		(void) strcpy(hostbuf, hp->h_name);
		return (hp->h_aliases);
	}
	else
		return (NULL);
}
/*
**  MAPHOSTNAME -- turn a hostname into canonical form
**
**	Parameters:
**		hbuf -- a buffer containing a hostname.
**		hbsize -- the size of hbuf.
**
**	Returns:
**		An exit code telling if the hostname was found and
**		canonicalized.
**
**	Side Effects:
**		Looks up the host specified in hbuf.  If it is not
**		the canonical name for that host, replace it with
**		the canonical name.  If the name is unknown, or it
**		is already the canonical name, leave it unchanged.
**/
bool
maphostname(hbuf, hbsize)
	char *hbuf;
	int hbsize;
{
	register struct hostent *hp;
	extern struct hostent *gethostbyname();
	static char tmphbuf[MAXNAME];
	extern int h_errno;
	int status;

	/*
	 * If first character is a bracket, then it is an address
	 * lookup.  Address is copied into a temporary buffer to
	 * strip the brackets and to preserve hbuf if address is
	 * unknown.
	 */
	if (*hbuf == '[')
	{
		extern struct hostent *gethostbyaddr();
		u_long in_addr;

		(void) strncpy(tmphbuf, hbuf+1, strlen(hbuf)-2);
		in_addr = inet_addr(tmphbuf);
		hp = gethostbyaddr(	(char *) &in_addr,
					sizeof(struct in_addr), AF_INET
		);
		if (hp) {
			if (strlen(hp->h_name) >= hbsize)
				hp->h_name[hbsize - 1] = '\0';
			(void) strcpy(hbuf, hp->h_name);
			return TRUE;
		}
		return FALSE;
	}

	status = getcanonname(hbuf, hbsize);
	if (tTd(9, 1))
	    printf("maphostname [getcanonname(%s,%d)] => %.*s\n",
		   hbuf, hbsize, hbsize-1,
		   (status == FALSE) ? "NOT_FOUND" : hbuf);
	return (status == FALSE) ? FALSE : TRUE;
}

# else DAEMON
/* code for systems without sophisticated networking */

/*
**  MYHOSTNAME -- stub version for case of no daemon code.
**
**	Can't convert to upper case here because might be a UUCP name.
**
**	Mark, you can change this to be anything you want......
*/

char **
myhostname(hostbuf, size)
	char hostbuf[];
	int size;
{
	register FILE *f;

	hostbuf[0] = '\0';
	f = fopen("/usr/include/whoami", "r");
	if (f != NULL)
	{
		(void) fgets(hostbuf, size, f);
		fixcrlf(hostbuf, TRUE);
		(void) fclose(f);
	}
	return (NULL);
}

/*
**  MAPHOSTNAME -- turn a hostname into canonical form
**
**	Parameters:
**		hbuf -- a buffer containing a hostname.
**		hbsize -- the size of hbuf.
**
**	Returns:
**		An exit code telling if the hostname was found and
**		canonicalized.
**
**	Side Effects:
**		Looks up the host specified in hbuf.  If it is not
**		the canonical name for that host, replace it with
**		the canonical name.  If the name is unknown, or it
**		is already the canonical name, leave it unchanged.
*/

/*ARGSUSED*/
maphostname(hbuf, hbsize)
	char *hbuf;
	int hbsize;
{
	return (FALSE);
}

#endif DAEMON

#ifdef IDA
/*
**  MAPINIT -- Open and (re)initialize a dbm database
**
**	Parameters:
**		c -- the (one character) name of the database
**
**	Returns:
**		An exit code telling if we could open the database.
**
*/
#ifdef NDBM
bool
mapinit(c)
    char c;
{
    struct stat stb;
    struct dbm_table *db;
    char buf[MAXNAME];

    db = &DbmTab[c & 0177];

    if (db->db_name == NULL) {
	syserr("database '%c' has not been defined", c);
	return FALSE;
    }

#ifdef YPMARK
    /*
     * Yellow pages are always supposed to be ready
     */
    if (db->db_name[0] == YPMARK)
	return TRUE;
#endif YPMARK

    /*
     * Have we already (unsuccessfully) tried to open it?
     */
    if (db->db_dbm == DB_NOSUCHFILE) {
	if (tTd(60, 1))
	    printf("mapinit(%c) => NO_FILE\n", c);
	return FALSE;
    }

    /*
     * If it already is open, check if it has been changed.
     */
    (void) sprintf(buf, "%s%s", db->db_name, DB_DIREXT);
    if (db->db_dbm != DB_NOTYETOPEN) {
	if (stat(buf, &stb) < 0 && (sleep(30), stat(buf, &stb) < 0)) {
	    syserr("somebody removed %s for db '%c'", buf, c);
	    db->db_dbm = DB_NOSUCHFILE;
	    if (tTd(60, 1))
		printf("mapinit(%c) => FILE_REMOVED\n", c);
	    return FALSE;
	}
	if (db->db_mtime != stb.st_mtime) {
	    if (tTd(60, 1))
		printf("database '%c' [%s] has changed; reopening it\n",
		       c, db->db_name);
	    (void) dbm_close(db->db_dbm);
	    db->db_dbm = DB_NOTYETOPEN;
	}
    }

    /*
     * Initialize database if not already open (r/w for aliases)
     */
    if (db->db_dbm == DB_NOTYETOPEN) {
	db->db_dbm = dbm_open(db->db_name,
			      c == DB_ALIAS ? O_RDWR : O_RDONLY, 0);
	if (db->db_dbm == DB_NOSUCHFILE) {
	    /* try once more */
	    sleep(30);
	    db->db_dbm = dbm_open(db->db_name,
				  c == DB_ALIAS ? O_RDWR : O_RDONLY, 0);
	}
	if (db->db_dbm == DB_NOSUCHFILE) {
	    syserr("can't open database '%c' [%s]", c, db->db_name);
	    if (tTd(60, 1))
		printf("mapinit(%c) => CAN'T OPEN %s\n", c, db->db_name);
	    return FALSE;
	}
	if (stat(buf, &stb) < 0 && (sleep(30), stat(buf, &stb) < 0)) {
	    syserr("can't stat %s", buf);
	    if (tTd(60, 1))
		printf("mapinit(%c) => FILE_REMOVED\n", c);
	    return FALSE;
	}
	db->db_mtime = stb.st_mtime;

	/*
	 * Make sure the database isn't being updated
	 */
	if (flock(dbm_dirfno(db->db_dbm), LOCK_EX | LOCK_NB) < 0)
	    if (errno == EWOULDBLOCK) {
		if (tTd(60, 1))
		    printf("%s%s is locked, waiting...\n",
			   db->db_name, DB_DIREXT);
		(void) flock(dbm_dirfno(db->db_dbm), LOCK_EX);
	    } else
		syserr("flock failed for db %c [%s]", c, db->db_name);
	(void) flock(dbm_dirfno(db->db_dbm), LOCK_UN);
    }
    return TRUE;
}
/*
**  MAPKEY -- Search a dbm database.
**
**	Search the named database using the given key.  If
**	a result is found, sprintf the argument through the
**	result back into the key and return TRUE;
**	otherwise return FALSE and do nothing.
**
**	Keysize may also be given as zero, in which case the
**	sprintf'ed result is returned if the key matched.
**
**	Parameters:
**		c -- the database
**		key -- search string
**		argval -- sprintf argument & result
**		argsiz -- size of argval
**
**	Returns:
**		An exit code telling if there was a match.
**
**	Side Effects:
**		The argval is rewritten to reflect what was found
**		in the database.
*/

mapkey(c, key, keysiz, arg)
    char c, *key, *arg;
    int keysiz;
{
    struct dbm_table *db;
    DATUM dkey, result;
    static char lowkey[MAXNAME];
#ifdef YPMARK
    static char *yp_domain = NULL;
#endif YPMARK

    db = &DbmTab[c & 0177];

    if (tTd(60, 1))
	printf("mapkey('%c', \"%s\", \"%s\") => ", c, key, arg ? arg : "--");

    /*
     * Init the database; return if failure
     */
    if (!mapinit(c))
	return FALSE;

    /*
     * Normalize key (ie turn it to lowercase)
     */
    (void) strcpy(lowkey, key);
    (void) makelower(lowkey);

#ifdef YPMARK
    /*
     * Test for yellow page database first
     */
    if (db->db_name[0] == YPMARK) {
	if (yp_domain == NULL)
	    (void) yp_get_default_domain(&yp_domain);

	/*
	 * We include the null after the string, but Sun doesn't
	 */
	if (yp_match(yp_domain, &db->db_name[1], lowkey,
		     strlen(key)+1, &result.dptr, &result.dsize) != 0 &&
	    yp_match(yp_domain, &db->db_name[1], lowkey,
		     strlen(key), &result.dptr, &result.dsize) != 0)
	    result.dptr = NULL;
	else
	    /* smash newline */
	    result.dptr[result.dsize] = '\0';
    } else {
#endif YPMARK
	/*
	 * Go look for matching dbm entry
	 */
	dkey.dptr = lowkey;
	dkey.dsize = strlen(dkey.dptr) + 1;
	result = dbm_fetch(db->db_dbm, dkey);
#ifdef YPMARK
    }
#endif YPMARK

    /*
     * Well, were we successful?
     */
    if (result.dptr == NULL) {
	if (tTd(60, 1))
	    printf("NOT_FOUND\n");
	return FALSE;
    }

    /*
     * Yes, rewrite result if sprintf arg was given.
     */
    if (arg == NULL)
	(void) strcpy(lowkey, result.dptr);
    else
	(void) sprintf(lowkey, result.dptr, arg);
    /* if keysiz is zero, that means we should return a string from the heap */
    if (keysiz == 0)
	key = newstr(lowkey);
    else {
	if (strlen(lowkey)+1 > keysiz) {
	    syserr("mapkey: result \"%s\" too long after expansion\n",
		   lowkey, keysiz);
	    lowkey[keysiz-1] = '\0';
	}
	(void) strcpy(key, lowkey);
    }
    if (tTd(60, 1))
	printf("%s\n", key);

    /* Ugly kludge that assumes that sizeof(int) == sizeof(char *) */
    return (int) key;
}

#else NDBM

/* should really read the table into the stab instead */
mapkey(db, key, keysiz, arg)
    char db, *key, *arg;
    int keysiz;
{
    return FALSE;
}

#endif NDBM
#endif IDA
